import matplotlib
import matplotlib as mpl
import pandas as pd
from matplotlib import pyplot as plt

from cumulative.options import options


class Figure:
    def __init__(self, x_label="X", y_label="Y", ioff=False):
        mpl.rcParams.update(mpl.rcParamsDefault)
        plt.rcParams["font.family"] = "monospace"
        cmap = matplotlib.colormaps["cool"]

        if ioff:
            plt.ioff()
        else:
            plt.ion()

        fig, ax = plt.subplots(figsize=(5, 5))

        ax.clear()
        ax.set_facecolor("black")
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        lim_pad = 0.05
        ax.set_xlim(0 - lim_pad, 1 + lim_pad)
        ax.set_ylim(0 - lim_pad, 1 + lim_pad)

        self.fig = fig
        self.ax = ax
        self.cmap = cmap


class Plot:
    def __init__(self, c):
        self.c = c

    def xrays(self, src=None, alpha=1, ms=1, lw=1, k=60, style=".", show=True):
        src = options.default_if_null(src, "transforms.source")
        tmp = options.get("transforms.tmp")
        with options.option_context({"transforms": {"source": tmp, "destination": tmp}}):
            self.c.fit(src=src, method="pchip3", k=k, n_samples=k)
            self.c.plot.scatter(alpha=alpha, ms=ms, lw=lw, style=style)
            self.c.drop()

    def heatmap(self, src=None, ms=4, k=60, score="idx", style=".", show=True):
        src = options.default_if_null(src, "transforms.source")
        tmp = options.get("transforms.tmp")
        tmp_score = f"{tmp}.score"
        with options.option_context({"transforms": {"source": tmp, "destination": tmp}}):
            self.c.fit(src=src, method="pchip3", k=k, n_samples=k)
            self.c.score(src=score, dst=tmp_score).sort(by=tmp_score)
            self.c.plot.scatter(ms=ms, alpha=tmp_score, score=tmp_score, style=style)
            self.c.drop()

    def highways(self, src=None, ms=4, k=60, score="idx", style=".", show=True):
        src = options.default_if_null(src, "transforms.source")
        tmp = options.get("transforms.tmp")
        tmp_score = f"{tmp}.score"
        with options.option_context({"transforms": {"source": tmp, "destination": tmp}}):
            # TODO: move sort to internals of scatter(), without sorting c.df
            self.c.score(src=score, dst=tmp_score).sort(by=tmp_score)
            self.c.plot.scatter(src=src, alpha=1, score=tmp_score, lw=1, style="-")
            self.c.drop()

    def pixelate(self, src=None, k=30, alpha=0.05, ms=5, style="s", show=True):
        src = options.default_if_null(src, "transforms.source")
        tmp = options.get("transforms.tmp")
        with options.option_context({"transforms": {"source": tmp, "destination": tmp}}):
            self.c.interp(n_samples=k, src=src)
            self.c.bin(n_bins=k)
            self.c.fit(method="pchip3", k=k, n_samples=k)
            self.c.plot.scatter(alpha=alpha, ms=ms, style=style)
            self.c.drop()

    def scatter(
        self, src=None, style=".", ms=2, lw=1, score=None, alpha=0.5, figure=None, x_label="x", y_label="y", show=True
    ):

        src = options.default_if_null(src, "transforms.source")

        if figure is None:
            figure = Figure(x_label=x_label, y_label=y_label)

        for _, row in self.c.df.iterrows():
            row_color = figure.cmap(row[score]) if isinstance(score, str) else "white"
            row_alpha = row[alpha] if isinstance(alpha, str) else alpha
            pd.Series(row[f"{src}.y"], index=row[f"{src}.x"]).plot(
                style=style,
                lw=lw,
                ms=ms,
                color=row_color,
                alpha=row_alpha,
                ax=figure.ax,
            )

        if show and plt.isinteractive():
            plt.show()



import logging

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from cumulative.options import options

log = logging.getLogger(__name__)


def process_row(func, row, **kwargs):
    return func(row, **kwargs)


class Transform:
    def __init__(self, c, name=None):
        self.c = c
        self.name = self.__class__.__name__.lower() if not name else name

    def transform_row(self, row):
        return pd.Series()

    def __call__(self, **kwargs):
        tqdm_params = options.get("tqdm.params")
        tqdm_params["desc"] = self.name
        tqdm.pandas(**tqdm_params)
        # The destination prefix is not required/expected by row transforms,
        # let's drop it if present.

        dst = options.default_if_null(kwargs.pop("dst", None), "transforms.destination")

        kwargs["src"] = options.default_if_null(kwargs.pop("src", None), "transforms.source")

        df = self.apply(**kwargs)

        self.c.track(self.name, dst, kwargs)

        if df is None:
            # Sort
            return self.c

        if isinstance(df, pd.DataFrame):
            df.columns = [f"{dst}.{col}" if col != "idx" else col for col in df.columns]
        elif isinstance(df, pd.Series):
            df = df.rename(dst).to_frame()
        else:
            raise Exception("Invalid argument type")

        self.validate(df)

        drop_cols = self.c.columns_with_prefix(dst, errors="ignore")
        self.c.df = self.c.df.drop(columns=drop_cols, errors="ignore")
        self.c.df = pd.concat(
            [self.c.df, df],
            axis=1,
        )
        return self.c

    def apply(self, **kwargs):
        return self.c.df.progress_apply(lambda row: self.transform_row(row, **kwargs), axis=1)

    def validate(self, df):
        # Check for invalid values (nans, infs)
        count_rows = len(df)
        count_invalid_rows = df.isin([np.inf, -np.inf, np.nan]).any(axis=1).sum()
        if count_invalid_rows > 0:
            log.warning(
                f"{self.name} produced {count_invalid_rows} ({count_invalid_rows / count_rows * 100:.0f}%)"
                " invalid rows (nans or infs)"
            )

        if len(set(df.columns)) != len(df.columns):
            log.warning(f"{self.name} produced duplicate column names")

        if len(df.columns) == 0:
            log.warning(f"{self.name} produced no columns")

        if len(df) == 0:
            log.warning(f"{self.name} produced no rows")



import copy
from contextlib import contextmanager

default_options = {
    "reproducibility": {"random_seed": 123},
    "tqdm": {"params": {"disable": False, "leave": False, "delay": 0}},
    "transforms": {
        "destination": "base",
        "source": "base",
        "tmp": "tmp",
        "sequence": {"attributes": "attrs"},
        "fit": {"curve_fit": {"maxfev": 10000}},
    },
    "doc": {"url": "https://elehcimd.github.io/cumulative/"},
}


class Options:
    _instance = None

    def __init__(self):
        raise RuntimeError("Call instance() instead")

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls._instance.options = copy.deepcopy(default_options)
        return cls._instance

    def default_if_null(self, value, path):
        if value is not None:
            return value
        else:
            return self.get(path)

    def get(self, path):
        steps = path.split(".")

        d = self.options

        for step in steps:
            d = d[step]
        return d

    def set(self, path, value):
        *steps, last = path.split(".")

        d = self.options

        for step in steps:
            d = d.setdefault(step, {})
        d[last] = value

    def reset(self, path=None):
        if path is None:
            self.options = copy.deepcopy(default_options)
            return

        steps = path.split(".")
        d = default_options
        for step in steps:
            d = d[step]
        self.set(path, d)

    @contextmanager
    def option_context(self, options):
        try:
            orig_options = copy.deepcopy(self.options)
            for k, v in options.items():
                self.set(k, v)
            yield self
        finally:
            self.options = orig_options


# This object handles the options for the package, process-wide.
# To change locally the preferences, use option_context.
options = Options.instance()



import logging

import pandas as pd

from cumulative.transforms.transform import Transform

log = logging.getLogger(__name__)


class ExceptionScaler(Exception):
    pass


class Scale(Transform):
    """
    MinMax normalizer on x,y axes to the unit interval [0,1].
    If the sequence contains less than two distinct values, all values are set to 1.
    """

    def transform_row(self, row, src):

        s = pd.Series(row[f"{src}.y"], index=row[f"{src}.x"])

        if len(s) == 0:
            raise ExceptionScaler(f"idx={row['idx']}: series length is zero")

        attrs = {
            "x_min": s.index.min(),
            "x_max": s.index.max(),
            "y_min": s.min(),
            "y_max": s.max(),
        }

        if len(s) == 1:
            log.warning(f"idx={row['idx']}: series length is below 2 (normalized to unit)")
            # Using 1 as default normalised value, to have integrals summing up to 1.
            s[:] = 1
        elif s.nunique() < 2:
            log.warning(f"idx={row['idx']}: series contains less than 2 distinct values (normalized to unit)")
            # Using 1 as default normalised value, to have integrals summing up to 1.
            s[:] = 1
        else:
            s = s - s.min()
            s = s / s.max()

        s.index -= s.index.min()
        s.index /= s.index.max()

        attrs = {
            **attrs,
            **{
                "x": s.index.values,
                "y": s.values,
            },
        }

        return pd.Series(attrs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cumulative', 'cumulative.transforms']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.1',
 'pandas>=1.5.3',
 'scikit-learn>=1.2.2',
 'scipy>=1.11.4',
 'statsmodels>=0.14.1',
 'tqdm>=4.66.1',
 'tsfel>=0.1.6']

setup_kwargs = {
    'name': 'cumulative',
    'version': '0.0.26',
    'description': 'Manipulation and analysis of data series collections',
    'long_description': '<p align="center">\n<img width="50%" height="50%" src="https://elehcimd.github.io/cumulative/assets/img/logo-black.svg" alt="CUMULATIVE Logo">\n</p>\n\n<p align="center">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/test.svg" alt="Test">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/coverage.svg" alt="Coverage">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/python.svg" alt="Python">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/pypi.svg" alt="PyPi">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/license.svg" alt="License">\n<img src="https://elehcimd.github.io/cumulative/assets/img/badges/code-style.svg" alt="Code style">\n</p>\n\n---\n\n**Cumulative** is an open source project and Python package that provides an intuitive in-memory interface to transform and analyse large collections of time series with composable pipelines.\n\n---\n\n* **Documentation**: [https://elehcimd.github.io/cumulative/](https://elehcimd.github.io/cumulative/)\n* **Source code**: [https://github.com/elehcimd/cumulative](https://github.com/elehcimd/cumulative)\n\n---\n\n## Key features\n\n* **Transformations**: Apply vectorized time series transformations as composable pipelines\n* **Interoperability**: Access the underlying data at any time as a Pandas data frame\n* **Flexibility**: Implement custom transformations and store arbitrary metadata\n\n## Requirements\n\n* **Python >=3.10**\n\n## Installation\n\n```\npip install cumulative\n```\n\n## License\n\nThis project is licensed under the terms of the [BSD 3-Clause License](https://elehcimd.github.io/cumulative/license/).\n\n\n',
    'author': 'Michele Dallachiesa',
    'author_email': 'michele.dallachiesa@sigforge.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://elehcimd.github.io/cumulative/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10.0',
}


setup(**setup_kwargs)



import pandas as pd

from cumulative.transforms.transform import Transform


class Select(Transform):
    def transform_row(self, row, src, func=lambda x: x):

        s = pd.Series(row[f"{src}.y"], index=row[f"{src}.x"])
        s = func(s)

        return pd.Series({"x": s.index.values, "y": s.values})
